//*****************************************************************************
//*****************************************************************************
//  FILENAME:   sensor.c
//  Version: 1.0, 2007/03/08
//  Generated by Andy Kiser
//
//  DESCRIPTION:  sensor specific functions 
//
//-----------------------------------------------------------------------------
// Originally authored by Toradex AG. This work is now available under the terms and conditions
// of the Creative Commons License 'Attribution CC BY', details of which can be found here:
// http://creativecommons.org/licenses/by/3.0/
//*****************************************************************************
//*****************************************************************************
//#define __DEBUG__
#include <m8c.h>        // part specific constants and macros
#include "sensor.h"
#include "psocgpioint.h"
#include "spim.h"
#include "feature.h"
#include "parameters.h"
#include "types_n_macros.h"
#include "arith.h"
//----------------------------------------
// Sensor parameters
// Factory Defaults are defined here (NUM_xxx are defined in "sensor.h")
//----------------------------------------
const BYTE FactoryDefault_Param_8   [NUM_PARAM_8] =
{
	0x00,		// report mode
	0x00,		// LED mode
	0x00		// SCA3000's operating MODE register
};

const WORD FactoryDefault_Param_16  [NUM_PARAM_16] = 
{
	100,		// report rate [ms]
	100,		// sample rate [ms]
	50000,		// sample rate [us]	(not used for this sensor)
	0,			// Offset X-Axis
	0,			// Offset Y-Axis
	0,			// Offset Z-Axis
};

const DWORD FactoryDefault_Param_32  [NUM_PARAM_32] =
{
	0x00000000	// currently no 32bit parameters
};

const tParam_168 FactoryDefault_Param_168 [NUM_PARAM_168] =
{
	"Accelerometer",	// user device name
	"Frame Number",		// user name Channel 0
	"Acceleration x",	// user name Channel 1
	"Acceleration y",	// user name Channel 2
	"Acceleration z" 	// user name Channel 3
};

//----------------------------------------
// SPI addresses and constants
//----------------------------------------
#define ADDR_ACC_X_LSB	0x04
#define ADDR_ACC_X_MSB	0x05
#define ADDR_ACC_Y_LSB	0x06
#define ADDR_ACC_Y_MSB	0x07
#define ADDR_ACC_Z_LSB	0x08
#define ADDR_ACC_Z_MSB	0x09
#define ADDR_MODE		0x14

//----------------------------------------
// this module's global variables
//----------------------------------------
tReport Report;

//----------------------------------------
// constant lookup-table for ADC->physical value conversion
//----------------------------------------
const int ADC2Acc_MSB[16] =	// Bits 12..9
{		
	0,
	1883,
	3766,
	5649,
	7532,
	9415,
	11298,
	13181,
	15064,
	16946,
	18829,
	20712,
	22595,
	24478,
	26361,
	28244	
};		
		
const int ADC2Acc_LSB[256] =	// Bits 0..7
{		
	0,
	7,
	15,
	22,
	29,
	37,
	44,
	51,
	59,
	66,
	74,
	81,
	88,
	96,
	103,
	110,
	118,
	125,
	132,
	140,
	147,
	154,
	162,
	169,
	177,
	184,
	191,
	199,
	206,
	213,
	221,
	228,
	235,
	243,
	250,
	257,
	265,
	272,
	279,
	287,
	294,
	302,
	309,
	316,
	324,
	331,
	338,
	346,
	353,
	360,
	368,
	375,
	382,
	390,
	397,
	405,
	412,
	419,
	427,
	434,
	441,
	449,
	456,
	463,
	471,
	478,
	485,
	493,
	500,
	508,
	515,
	522,
	530,
	537,
	544,
	552,
	559,
	566,
	574,
	581,
	588,
	596,
	603,
	610,
	618,
	625,
	633,
	640,
	647,
	655,
	662,
	669,
	677,
	684,
	691,
	699,
	706,
	713,
	721,
	728,
	736,
	743,
	750,
	758,
	765,
	772,
	780,
	787,
	794,
	802,
	809,
	816,
	824,
	831,
	838,
	846,
	853,
	861,
	868,
	875,
	883,
	890,
	897,
	905,
	912,
	919,
	927,
	934,
	941,
	949,
	956,
	964,
	971,
	978,
	986,
	993,
	1000,
	1008,
	1015,
	1022,
	1030,
	1037,
	1044,
	1052,
	1059,
	1067,
	1074,
	1081,
	1089,
	1096,
	1103,
	1111,
	1118,
	1125,
	1133,
	1140,
	1147,
	1155,
	1162,
	1169,
	1177,
	1184,
	1192,
	1199,
	1206,
	1214,
	1221,
	1228,
	1236,
	1243,
	1250,
	1258,
	1265,
	1272,
	1280,
	1287,
	1295,
	1302,
	1309,
	1317,
	1324,
	1331,
	1339,
	1346,
	1353,
	1361,
	1368,
	1375,
	1383,
	1390,
	1397,
	1405,
	1412,
	1420,
	1427,
	1434,
	1442,
	1449,
	1456,
	1464,
	1471,
	1478,
	1486,
	1493,
	1500,
	1508,
	1515,
	1523,
	1530,
	1537,
	1545,
	1552,
	1559,
	1567,
	1574,
	1581,
	1589,
	1596,
	1603,
	1611,
	1618,
	1626,
	1633,
	1640,
	1648,
	1655,
	1662,
	1670,
	1677,
	1684,
	1692,
	1699,
	1706,
	1714,
	1721,
	1728,
	1736,
	1743,
	1751,
	1758,
	1765,
	1773,
	1780,
	1787,
	1795,
	1802,
	1809,
	1817,
	1824,
	1831,
	1839,
	1846,
	1854,
	1861,
	1868,
	1876	
};		


//----------------------------------------
// type definitions
//----------------------------------------

typedef enum 
{	
	eReset,
	eIdle,
	eReadAcceleration,
	eTransmitData
} tFSM_Measure;

typedef struct
{
	union uBigEndian Axis[3];
} tWordVector3D;


//----------------------------------------
// variables
//----------------------------------------
tFSM_Measure	FSM;
WORD			Countdown_1ms;

BOOL			NewSample = FALSE;

// store original drivemodes during suspend
BYTE			DM0_2, DM0_1, DM0_0;
BYTE			DM2_2, DM2_1, DM2_0;
BYTE			Mode;
extern BYTE		SOF_Flag; // needed to compensate for FSM processing duration >1ms
//----------------------------------------
// function prototypes
//----------------------------------------
static WORD sqrt(DWORD arg);
static int Calculate_g(union uBigEndian ADC);
static WORD Atan2 (int x, int y);

//-----------------------------------------------------------------------------
// Initializes the Sensor 
//-----------------------------------------------------------------------------
void Sensor_Start ()
{
	FSM = eReset;

	// Initialize SPI to mode 3 as it is neccessary to communicate with the SMB365 sensor
	SPIM_Start(SPIM_SPIM_MODE_0 | SPIM_SPIM_MSB_FIRST);
	// nCS idle state = 1
	GPIO_LOW(nRST)
	GPIO_HIGH(nCS)
	GPIO_HIGH(nRST)
	
}
//-----------------------------------------------------------------------------
// Prepare Sensor for Suspend
//-----------------------------------------------------------------------------
void Sensor_Suspend()
{
	// put SCA3000 into Reset mode
	GPIO_LOW(nRST)
		
	// store original drivemodes during  suspend
	DM0_2 = PRT0DM2;
	DM0_1 = PRT0DM1;
	DM0_0 = PRT0DM0;
	DM2_2 = PRT2DM2;
	DM2_1 = PRT2DM1;
	DM2_0 = PRT2DM0;
	
	// do NOT set all I/Os to/from the sensor to High-Z Analog (=110), 
	// undefined I/Os may lead to excessive current drain in standby mode
/*	PRT0DM2  = 0xFF;
	PRT0DM1  = 0xFF;
	PRT0DM0  = 0x00;
	PRT2DM2  = 0xFF;
	PRT2DM1  = 0xFF;
	PRT2DM0  = 0x00;
*/
}
//-----------------------------------------------------------------------------
// Resume Sensor from Suspend mode
//-----------------------------------------------------------------------------
void Sensor_Resume()
{
	// restore drivemodes for GPIOs to/from the sensor
	PRT2DM2 = DM2_2;
	PRT2DM1 = DM2_1;
	PRT2DM0 = DM2_0;

	PRT0DM2 = DM0_2;
	PRT0DM1 = DM0_1;
	PRT0DM0 = DM0_0;

	// start up sensor again
	Sensor_Start();
}

//-----------------------------------------------------------------------------
// Read one register byte from the sensor
// The function assumes the following (idle) conditions on entry:
// 		CSn == 1
// 		SPI-Mode == 0
//-----------------------------------------------------------------------------
BYTE Sensor_ReadByte(BYTE Address)
{
	GPIO_LOW(nCS)

	SPIM_bReadRxData(); 	  // -- dummy read to clear RX_BUFFER_FULL flag
	SPIM_SendTxData(Address << 2); // Transmit Address
	while( !( SPIM_bReadStatus() & SPIM_SPIM_RX_BUFFER_FULL ) );   	
	SPIM_bReadRxData(); 			// -- dummy read to clear RX_BUFFER_FULL flag

	// -- Receive Data Byte
	SPIM_SendTxData(0x00);
	while( !( SPIM_bReadStatus() & SPIM_SPIM_RX_BUFFER_FULL ) );   	
	
	GPIO_HIGH(nCS)
	
	return SPIM_bReadRxData();
}

//-----------------------------------------------------------------------------
// Read one register byte from the sensor
// The function assumes the following (idle) conditions on entry:
// 		CSn == 1
// 		SPI-Mode == 0
//-----------------------------------------------------------------------------
void Sensor_WriteByte(BYTE Address, BYTE Value)
{
	GPIO_LOW(nCS)

	SPIM_SendTxData((Address << 2) |0x02); // Transmit Address
	while( !( SPIM_bReadStatus() & SPIM_SPIM_TX_BUFFER_EMPTY ) );

	// -- Transmit Data Byte
	SPIM_SendTxData(Value);
	while( !( SPIM_bReadStatus() & SPIM_SPIM_TX_BUFFER_EMPTY ) );
	
	GPIO_HIGH(nCS)
}

//-----------------------------------------------------------------------------
// Read Raw ADC value from the sensor
// The function assumes the following (idle) conditions on entry:
// 		CSn == 1
// 		SPI-Mode == 0
//-----------------------------------------------------------------------------
void Sensor_ReadAcceleration (tWordVector3D *Acceleration)
{
	tWordVector3D Acc;

	union uBigEndian g_raw;
	signed short xyz;

	GPIO_LOW(nCS)

	SPIM_bReadRxData(); 			 // -- dummy read to clear RX_BUFFER_FULL flag
	SPIM_SendTxData(ADDR_ACC_Z_MSB << 2); // Transmit Address
	while( !( SPIM_bReadStatus() & SPIM_SPIM_RX_BUFFER_FULL ) );   	
	SPIM_bReadRxData(); 			// -- dummy read to clear RX_BUFFER_FULL flag

	for (xyz=2; xyz>=0; xyz--) 
	{		
		// -- Receive Data Byte
		SPIM_SendTxData(0x00);
		while( !( SPIM_bReadStatus() & SPIM_SPIM_RX_BUFFER_FULL ) );   	
		g_raw.BE.Hi = SPIM_bReadRxData();
		Acc.Axis[xyz].BE.Hi = SPIM_bReadRxData();
	
		// -- Receive Data Byte
		SPIM_SendTxData(0x00);
		while( !( SPIM_bReadStatus() & SPIM_SPIM_RX_BUFFER_FULL ) );   	
		//g_raw.BE.Lo |= SPIM_bReadRxData(); // Error versions <1.4
		g_raw.BE.Lo = SPIM_bReadRxData(); 

		// -- shift into correct position
		g_raw.i >>= 3;

		// -- handle overflow / underflow of g-range
		if ((signed char)g_raw.BE.Hi >= 3072/256) 	// 0000'1100'0000'0000b -> positive g overflow
			 g_raw.i = 3071;
		else 
			if ((signed char)g_raw.BE.Hi < -3072/256) 	// 1111'0011'1111'1111b -> negative overflow
				g_raw.i = -3072;

		Acceleration->Axis[xyz&0x03] = g_raw; // sort axes in the correct order
	}	
	
	GPIO_HIGH(nCS)
}


//-----------------------------------------------------------------------------
// Function that will be called when a FEATURE REPORT is received that 
// cannot be handled with the standard functionality in feature.c
// The feature report data is available in the "FEATURE_RPT" variable
//
// --> see feature.c for an example how to parse FEATURE_RPT
//
//	return Value: number of Parameter bytes (1-31) that are 
//				  returned in the IN report
//-----------------------------------------------------------------------------
BYTE Sensor_Specific_Feature(void)
{
	BYTE i;
	
	static const BYTE ValidRegister[] = 
	{
		0x00, // REVID
		0x02, // STATUS
		0x04, // X_LSB
		0x05, // X_MSB
		0x06, // Y_LSB
		0x07, // Y_MSB
		0x08, // Z_LSB
		0x09, // Z_MSB
		0x0F, // BUF_DATA
		0x12, // TEMP_LSB
		0x13, // TEMP_MSB
		0x14, // MODE
		0x15, // BUF_COUNT
		0x16, // INT_STATUS
		0x17, // I2C_RD_SEL
		0x18, // CTRL_SEL
		0x1E, // UNLOCK
		0x21, // INT_MASK
		0x22  // CTRL_DATA
	};
	
	if ( (FEATURE_RPT.OUT.Target == eSensor) && (FEATURE_RPT.OUT.ParamSize == 1) )
		for (i=0; i<sizeof(ValidRegister); i++)
			if (FEATURE_RPT.OUT.Index_LSB == ValidRegister[i])
				if (FEATURE_RPT.OUT.Get_nSet)
				{ // GET  -> Read a random SCA3000 register
					FEATURE_RPT.IN.Param[0] = Sensor_ReadByte(FEATURE_RPT.OUT.Index_LSB);
					return 1;
				}
				else
				{ // SET  -> Write a random SCA3000 register
					Sensor_WriteByte(FEATURE_RPT.OUT.Index_LSB, FEATURE_RPT.OUT.Param[0]);
					return 0;
				}
	return 0;	
}

//-----------------------------------------------------------------------------
// Function that will be called when a FEATURE REPORT is received that 
// sets / gets a 8/16/32/168bit parameter in feature.c
// The feature report data is available in the "FEATURE_RPT" variable
//
// --> see feature.c for an example how to parse FEATURE_RPT
//-----------------------------------------------------------------------------
void Sensor_Specific_Parameter(BYTE *NumberOfReturnedBytes)
{
	// currently there are no specific parameters
}

//-----------------------------------------------------------------------------
// void Sensor_FSM_Measure(void)
//
//	finite state machine to control the measurement process for 
//  the Sensor
//
//  IMPORTANT: 	call this routine once every millisecond. Otherwise the timing
//				parameters cannot be met.
//-----------------------------------------------------------------------------
#define X 0
#define Y 1
#define Z 2

void Sensor_FSM_Measure() 
{
	tWordVector3D  		g_raw;
	union uBigEndian	Acceleration[3];
	//union uBigEndian	AccXY, AccTot, Theta, Phi;
	union uBigEndian	g, offset;
	//WORD				xabs, yabs, zabs;
	//long int 			x2, y2, z2, x2y2, x2y2z2;

    BYTE 				SCA3000_MODE_ToBe;				// Operating Mode that is set by the user
	static BYTE			SCA3000_MODE_Readback = 0x00;	// Operating Mode read back from the sensor
	
	// decrement millisecond counter
	if (Countdown_1ms != 0)
		Countdown_1ms--;

	// process actual state machine
	switch (FSM)
	{
		case eReset:
			SCA3000_MODE_Readback = Sensor_ReadByte(ADDR_MODE);
			Countdown_1ms = 35; // time for the sensor to stabilize after power-up
			FSM = eIdle;
			break;
		
		case eIdle:
			if (Countdown_1ms == 0)
			{
				// reset samplerate counter
				Get_Param_16(PAR16_SAMPLERATE_MS, VOLATILE, &Countdown_1ms);
				FSM = eReadAcceleration;
			}
			break;		

		case eReadAcceleration:
			Sensor_ReadAcceleration (&g_raw);
			
			// -- Fill in the results into IN report
			g.i = Calculate_g( g_raw.Axis[X] ); 
			Get_Param_16(PAR16_OFFSET_X, VOLATILE, &offset.w);
			g.i = g.i + offset.w;
			Report.Ch1_LSB = g.BE.Lo;
			Report.Ch1_MSB = g.BE.Hi;

			g.i = Calculate_g( g_raw.Axis[Y] ); 
			Get_Param_16(PAR16_OFFSET_Y, VOLATILE, &offset.w);
			g.i = g.i + offset.w;
			Report.Ch2_LSB = g.BE.Lo;
			Report.Ch2_MSB = g.BE.Hi;

			g.i = Calculate_g( g_raw.Axis[Z] ); 
			Get_Param_16(PAR16_OFFSET_Z, VOLATILE, &offset.w);
			g.i = g.i + offset.w;
			Report.Ch3_LSB = g.BE.Lo;
			Report.Ch3_MSB = g.BE.Hi;

/*			// -- calculate Total Acceleration
				// x^2 + y^2 + z^2
			mul_16x16_32(g_raw.Axis[X].i, g_raw.Axis[X].i, &x2y2);
			mul_16x16_32(g_raw.Axis[Y].i, g_raw.Axis[Y].i, &y2);
			x2y2 += y2;											 // store x^2+y^2 for Zenith calculation
			mul_16x16_32(g_raw.Axis[Z].i, g_raw.Axis[Z].i, &z2);
			x2y2z2 = x2y2 + z2;
			
			AccTot.w = sqrt(x2y2z2);
			AccTot.i = Calculate_g(AccTot);

			// -- calculate Azimuth
			Theta.w = Atan2(g_raw.Axis[X].i, g_raw.Axis[Y].i);
			
			// -- Calculate Zenith
			AccXY.w = sqrt(x2y2);
			Phi.w = Atan2(g_raw.Axis[Z].i, AccXY.w);

			// -- Fill in the results into IN report
			Report.Ch1_LSB = AccTot.BE.Lo;
			Report.Ch1_MSB = AccTot.BE.Hi;

			Report.Ch2_LSB = Theta.BE.Lo;
			Report.Ch2_MSB = Theta.BE.Hi;

			Report.Ch3_LSB = Phi.BE.Lo;
			Report.Ch3_MSB = Phi.BE.Hi;
*/
			FSM = eTransmitData;
			break;			
			
		case eTransmitData:
			// set new mode if requested
			Get_Param_8(PAR8_SCA3000_MODE, VOLATILE, &SCA3000_MODE_ToBe);
			
			if (SCA3000_MODE_Readback != SCA3000_MODE_ToBe)
			{
				Sensor_WriteByte(ADDR_MODE, SCA3000_MODE_ToBe);
				// read back sensor mode
				SCA3000_MODE_Readback = Sensor_ReadByte(ADDR_MODE);
			}

			// signal availability of new data
			NewSample = TRUE;
			
			FSM = eIdle;
			break;			
	}
	
	// compensate timing if processing above took longer than 1ms
	while (SOF_Flag >0)
	{
		SOF_Flag--;
		if (Countdown_1ms != 0)
			Countdown_1ms--;
	}
}
//-----------------------------------------------------------------------------
// calculate square root 32bit -> 16bit
// algorithm from
// http://www.ccp14.ac.uk/ccp/ccp14/ftp-mirror/nih-image/pub/nih-image/documents/square_root_code.txt
//-----------------------------------------------------------------------------
static WORD sqrt(DWORD arg)
{
        DWORD Rect, add, square;
 
        Rect = 0;
        square = 0x40000000;
        do {
                add = Rect + square;
                if (add <= arg) {
                        arg -= add;
                        Rect = add + square;
                }
                Rect >>= 1;
                square >>= 1;
                square >>= 1; // leads to smaller and faster code than "square >>= 2"
        } while (square != 0);
 
        if (arg > Rect)
                Rect++;
        return(Rect);
}
//-----------------------------------------------------------------------------
// calculate arc tan using vector math
// this routine provides valid results only for the first quadrant 0..pi/2 rad
// --> see Atan2() below
//-----------------------------------------------------------------------------
// lookup table: 1e4*{arctan(1/1), arctan(1/2), arctan(1/4), arctan(1/8), arctan(1/16), ...}
// last value is 1.5x the calculated value, for reasons of rounding.
static WORD Theta[12] = {7854, 4636, 2450, 1244, 624, 312, 156, 78, 39, 20, 10, 7};

static WORD Atan_Quadrant1 (WORD x, WORD y)
{
	WORD iAngle;
	union 
	{
		DWORD dw;
		BYTE  b[4];
	} Xval, Yval;
	DWORD Xtemp, Ytemp, Yshl;
	BYTE n;

	Xval.dw = x; 
	Yval.dw = y;
	iAngle = 0;

	// in each loop, do a complex multiplication
	//		(Xval|Yval) = (Xval|Yval) x (2^n|-1)
	// This rotates the vector clockwise by the angle of the vector (2^n|-1) --> see table Theta
	// if the multiplication rotates the vector too far (into quadrant IV), it is *not* performed.
	for (n=0; n<12; n++)
	{	
		// -- check if rotation would bring the vector into quadrant IV.
        Yshl = Yval.dw << n;
        if (Yshl >= Xval.dw)	
		{
			// -- update the sum by which the vector has been rotated (this is the result of the function!)
			iAngle += Theta[n];
			 
			// -- do the vector multiplication to perform the actual rotation
            // x.new = 2^n*x + y
            // y.new = 2^n*y - x
            Xtemp = Xval.dw;
            Ytemp = Yval.dw;
            Yval.dw = Yshl;
            Xval.dw <<= n;
            
            Xval.dw += Ytemp;
            Yval.dw -= Xtemp;
            
			// -- scale x and y for the next loop pass.
            while ( ((Xval.b[0] != 0) || (Xval.b[1] >= 0x10) ))	/// (Xval.dw > 1'048'575)
			{
				Xval.dw += 128;
				// Xval.dw /= 256;
				Xval.b[3] = Xval.b[2];
				Xval.b[2] = Xval.b[1];
				Xval.b[1] = Xval.b[0];
				Xval.b[0] = 0;
				
				Yval.dw += 128;
				// Yval.dw /= 256;
				Yval.b[3] = Yval.b[2];
				Yval.b[2] = Yval.b[1];
				Yval.b[1] = Yval.b[0];
				Yval.b[0] = 0;
			}
		}
	}
    
    return iAngle;
}
//-----------------------------------------------------------------------------
// calculate arc tan, preserving the correct quadrant
// return value is in the interval [0 ... 2pi] rad, scaled by 10'000
// --> this is [0 ... 62'831] 
//-----------------------------------------------------------------------------
static WORD Atan2 (int x, int y)
{
	if (x >= 0)
	{
		if (y == 0)
			return 0;
		if (y >= 0)
		{	// Quadrant I
			return Atan_Quadrant1(x, y);
		}
		else
		{	// Quadrant IV
			return Atan_Quadrant1(-y, x) + 47124;  // 47'124 = 1.5 x pi x 1e4 (270)
		}
	}
	else
	{
		if (y >= 0)
		{	// Quadrant II
			return Atan_Quadrant1(y, -x) + 15708;  // 15'708 = 0.5 x pi x 1e4 (90)
		}
		else
		{	// Quadrant III
			return Atan_Quadrant1(-x, -y) + 31416; // 31'416 =       pi x 1e4 (180)
		}
	}
}

//-----------------------------------------------------------------------------
// calculate acceleration, out of the raw ADC value
//-----------------------------------------------------------------------------

static int Calculate_g(union uBigEndian ADC)
{
	union uBigEndian Negative;
	
	if ((ADC.BE.Hi &0x80) == 0x0) // positive
		return ADC2Acc_MSB[ADC.BE.Hi] + ADC2Acc_LSB[ADC.BE.Lo];
	else						// negative
	{
		Negative.i = -ADC.i;
		return -( ADC2Acc_MSB[Negative.BE.Hi] + ADC2Acc_LSB[Negative.BE.Lo] );
	}
}

//-----------------------------------------------------------------------------
void Sensor_Debug(void)
{
/* currently no debug function */
}